"""utils.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""


class ProtectedList(list):

    def __init__(
            self,
            *args,
            protect=True):
        self._protect = protect
        super().__init__(*args)

    def __setitem__(self, key, value):
        if self._protect is True:
            raise TypeError('This list is protected.')
        elif callable(self._protect):
            self._protect(value)
        super().__setitem__(key, value)

    def append(self, p_object):
        if self._protect is True:
            raise TypeError('This list is protected.')
        elif callable(self._protect):
            self._protect(p_object)
        super().append(p_object)

    def extend(self, iterable):
        if self._protect is True:
            raise TypeError('This list is protected.')
        elif callable(self._protect):
            values = []
            for value in iterable:
                self._protect(value)
                values.append(value)
            iterable = iter(values)

        super().extend(iterable)

    def pop(self, index=None):
        if self._protect is True:
            raise TypeError('This list is protected.')
        return super().pop(index)

    def clear(self):
        if self._protect is True:
            raise TypeError('This list is protected.')
        return super().clear()

    def insert(self, index, p_object):
        if self._protect is True:
            raise TypeError('This list is protected.')
        elif callable(self._protect):
            self._protect(p_object)
        super().insert(index, p_object)

    def remove(self, value):
        if self._protect is True:
            raise TypeError('This list is protected.')
        return super().remove(value)

    def reverse(self):
        if self._protect is True:
            raise TypeError('This list is protected.')
        return super().reverse()

    def __add__(self, other):
        self.extend(other)
        return self

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __delitem__(self, key):
        if self._protect is True:
            raise TypeError('This list is protected.')
        return super().__delitem__(key)


if __name__ == '__main__':
    def check_int(val):
        if not isinstance(val, int):
            raise TypeError('Only integers are allowed.')

    l = [1, 2, 3]
    p = ProtectedList(l, protect=check_int) + [3]
    p.append(5)
    p.extend([6])
    p += [7]
    print(p)

    r = ProtectedList(l)
    del r[2]


"""path.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from .pathelement import PathElement
from .pathelement import path_element_from_decoder
from .buffer import BufferDecodeError


def path_from_decoder(decoder):
    pairs = []
    while decoder:
        tt = decoder.get_var_int32()
        if tt == 11:
            pairs.append(path_element_from_decoder(decoder))
            continue

        if tt == 0:
            raise BufferDecodeError('corrupted')

    return Path(pairs=pairs)


class Path:

    def __init__(self, pairs):
        self._path = tuple(
            pe if isinstance(pe, PathElement) else PathElement(*pe)
            for pe in pairs)

    def encode(self, buffer):
        buffer.add_var_int32(114)
        buffer.add_var_int32(self.byte_size)

        for path_element in self._path:
            buffer.add_var_int32(11)
            path_element.encode(buffer)
            buffer.add_var_int32(12)

    def __getitem__(self, item):
        return self._path.__getitem__(item)

    def __repr__(self):
        return str(tuple((pe.kind, pe.id) for pe in self._path))

    def get_dict(self):
        return {'path': [pe.get_dict() for pe in self._path]}

    @property
    def byte_size(self):
        n = 2 * len(self._path)
        for path_element in self._path:
            n += path_element.byte_size
        return n



"""__init__.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from .integervalue import IntegerValue
from .stringvalue import StringValue
from .booleanvalue import BooleanValue
from .arrayvalue import ArrayValue
from .jsonvalue import JsonValue
from .doublevalue import DoubleValue
from .keyvalue import KeyValue
from .anyvalue import AnyValue



"""entity.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import json
from .key import Key
from .timestampvalue import TimestampValue
from .utils import value_from_dict
from .utils import value_to_dict


class Entity:

    def __init__(self, entity_res):
        """Initialize an Entity object.

        Example:
        {
            'properties': {
                'name': {'stringValue': 'Example'},
                ...
            },
            'key': {
                'partitionId': {'projectId': 'my_project_id'},
                'path': [
                    {'kind': 'Foo', 'id': 1234},
                    ...
                ]
            }
        }

        See the following link for more information:
        https://cloud.google.com/datastore/docs/reference/rest/v1/Entity
        """
        self.key = Key(entity_res['key'])
        self._properties = set()

        for prop, val in entity_res['properties'].items():
            self._properties.add(prop)
            setattr(self, prop, value_from_dict(val))

    def __str__(self):
        return json.dumps(self.serializable_dict())

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def get_dict(self):
        """Returns dictionary object which can be used to insert, upsert or
        update the entity in the google cloud datastore."""

        # We use __dict__ instead of getattr() since the Entity class might be
        # sub-classed and then getattr() could access a computed property
        # instead of the variable we really want.

        return {
            'key': self.key.get_dict(),
            'properties': {
                prop: value_to_dict(self.__dict__[prop])
                for prop in self._properties
            }
        }

    def serializable_dict(self, key_as=None):
        data = {
            prop: _serialize_value(self.__dict__[prop])
            for prop in self._properties
        }
        if isinstance(key_as, str):
            data[key_as] = self.key.ks
        return data

    def set_property(self, prop, value):
        """Use this method to set a new or change an existing property.

        If you are sure the property already exists, its possible to set the
        property directly. This method must be used for new properties.
        """
        self.__dict__[prop] = value
        self._properties.add(prop)


def _serialize_value(val):
    if isinstance(val, TimestampValue):
        return str(val)
    if isinstance(val, Key):
        return val.ks
    if isinstance(val, list):
        return [_serialize_value(v) for v in val]
    return val



"""model.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import functools
from ..connector.entity import Entity
from ..orm.properties.value import Value
from ..connector.key import Key
from .filter import Filter
from ..connector.timestampvalue import TimestampValue


class _PropertyClass(dict):
    """Custom dictionary for a GcdModel."""

    def __init__(self, *args):
        self._props = {}
        super().__setitem__('model_props', self._props)

    def __setitem__(self, key, value):
        """Sets the name property for each orm property. Each property is
        added to the 'model_props' dictionary which is bound to the model
        class.
        """
        if isinstance(value, Value):
            value.name = key
            self._props[key] = value
        super().__setitem__(key, value)


class _ModelClass(type):
    """Used as a meta class and uses the __preprare__ method to return a custom
    class dictionary object.
    """
    @classmethod
    def __prepare__(mcs, name, bases):
        return _PropertyClass()

    def __new__(mcs, name, *args, **kwargs):
        result = super().__new__(mcs, name, *args, **kwargs)
        result.__class__.__name__ = name
        return result


class GcdModel(Entity, metaclass=_ModelClass):
    ALLOW_NEW_PROPERTIES = False
    BASE_MODEL_INIT = """
    Model can be initialized using an entity object or by using keyword
    arguments and a key object.

    Examples:
        MyModel(Entity(...))
        MyModel(name='foo', age=3, ..., key=Key(...))
"""
    __kind__ = None

    def __init__(self, entity=None, key=None, **template):
        """Initialize a GcdModel.

        You can initialize a model by using either an Entity OR a Key and
        template.

        :param entity: Entity object.
        :param key: Key object.
        :param template: keyword arguments defining the model properties.
        """
        props = set(self.model_props.keys())

        if entity is not None:
            assert key is None and len(template) == 0, self.BASE_MODEL_INIT
            self.__dict__.update(entity.__dict__)
            props -= set(self.__dict__.keys())

        else:
            assert isinstance(key, Key), self.BASE_MODEL_INIT
            self._properties = set()
            self.key = key

            for name, value in template.items():
                if name in props:
                    props.remove(name)
                    setattr(self, name, value)

        if self.__kind__ != self.key.kind:
            raise TypeError(
                'Expecting kind {expect!r} for model {model} but got {got!r}. '
                'Optionally you can set {model}.__kind__ to {got!r} or to '
                'None if you want to ignore the kind check.'
                .format(
                    expect=self.__kind__,
                    model=self.__class__.__name__,
                    got=self.key.kind))

        for prop in props:
            prop = self.model_props[prop]
            if prop.default is not None:
                setattr(
                    self,
                    prop.name,
                    prop.default() if callable(prop.default) else prop.default)
            elif prop.required:
                raise TypeError(
                    'Missing required property: {}'.format(prop.name))
            else:
                super().__setattr__(prop.name, None)

    def __new__(mcs, *args, **kwargs):
        if getattr(mcs, '__kind__') is None:
            mcs.__kind__ = mcs.__name__
        return super().__new__(mcs)

    def __getattribute__(self, key):
        if key != 'model_props' and key in self.model_props:
            return self.model_props[key].get_value(self)
        return super().__getattribute__(key)

    def __setattr__(self, key, value):
        if key in self.model_props:
            self.model_props[key].set_value(self, value)
        else:
            super().__setattr__(key, value)

    def set_property(self, prop, value):
        if prop in self.model_props:
            self.model_props[prop].set_value(self, value)
        elif not self.ALLOW_NEW_PROPERTIES:
            raise RuntimeError(
                'Adding new properties on this model is not allowed. '
                'Set {}.ALLOW_NEW_PROPERTIES to True if you really want add '
                'property {!r} to this model.'
                .format(self.__class__.__name__, prop))
        super().set_property(prop, value)

    @classmethod
    def filter(cls, *filters, has_ancestor=None, key=None):
        return Filter(
            cls,
            *filters,
            has_ancestor=has_ancestor,
            key=key)

    @classmethod
    def get_kind(cls):
        if cls.__kind__ is None:
            cls.__kind__ = cls.__name__
        return cls.__kind__

    @classmethod
    async def get_entities(cls, gcd):
        return await Filter(cls).get_entities(gcd)

    def serializable_dict(self, key_as=None):
        data = {
            prop.name: self._serialize_value(prop.get_value(self))
            for prop in self.model_props.values()
            if prop.get_value(self) is not None
        }

        if isinstance(key_as, str):
            data[key_as] = self.key.ks

        return data

    @classmethod
    def _serialize_value(cls, val):
        if isinstance(val, TimestampValue):
            return str(val)
        if isinstance(val, Key):
            return val.ks
        if isinstance(val, list):
            return [cls._serialize_value(v) for v in val]
        return val



from .buffer import BufferDecodeError

TYPE_ID = 0
TYPE_NAME = 1


def path_element_from_decoder(decoder):
    kind = None
    name_or_id = None

    while True:
        tt = decoder.get_var_int32()

        if tt == 12:
            break
        if tt == 18:
            kind = decoder.get_prefixed_string()
            continue
        if tt == 24:
            name_or_id = decoder.get_var_int64()
            continue
        if tt == 34:
            name_or_id = decoder.get_prefixed_string()
            continue
        if tt == 0:
            raise BufferDecodeError('corrupt')

    assert kind is not None and name_or_id is not None, \
        'Expecting a path element with a kind and name/id.'

    return PathElement(kind, name_or_id)


class PathElement:

    def __init__(self, kind, name_or_id):
        assert name_or_id is None or isinstance(name_or_id, (int, str)), \
            'Expecting a str or int type but got: {}'.format(
                type(name_or_id))
        self.kind = kind
        self.id = name_or_id

    @property
    def name(self):
        raise TypeError('Use .id instead of .name')

    def encode(self, buffer):
        buffer.add_var_int32(18)
        buffer.add_prefixed_string(self.kind)
        if isinstance(self.id, int):
            buffer.add_var_int32(24)
            buffer.add_var_int64(self.id)
        else:
            buffer.add_var_int32(34)
            buffer.add_prefixed_string(self.id)

    @property
    def byte_size(self):
        n = self._size_str(self.kind)
        if isinstance(self.id, int):
            n += 1 + self._size_var_int(self.id)
        else:
            n += 1 + self._size_str(self.id)

        return n + 1

    def get_dict(self):
        if isinstance(self.id, int):
            return {'kind': self.kind, 'id': str(self.id)}

        if isinstance(self.id, str):
            return {'kind': self.kind, 'name': self.id}

        return {'kind': self.kind}

    @classmethod
    def _size_str(cls, s):
        l = len(s)
        return cls._size_var_int(l) + l

    @staticmethod
    def _size_var_int(n):
        if n < 0:
            return 10

        result = 0
        while True:
            result += 1
            n >>= 7
            if n == 0:
                break
        return result



"""integervalue.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from .value import Value


class IntegerValue(Value):

    def check_value(self, value):
        if not isinstance(value, int):
            raise TypeError(
                'Expecting an value of type \'int\' for property {!r} '
                'but received type {!r}.'
                .format(self.name, value.__class__.__name__))

    def set_value(self, model, value):
        self.check_value(value)
        # wrap value so booleans are converted to int type
        super().set_value(model, int(value))



"""utils.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import base64
from .key import Key
from .timestampvalue import TimestampValue

# Max string length in bytes. (strings are converted to blob values when larger
# than _MAX_STRING_LENGTH)
_MAX_STRING_LENGTH = 1500


def value_to_dict(val):
    if val is None:
        return {'nullValue': None}
    if isinstance(val, bool):
        return {'booleanValue': val}
    if isinstance(val, int):
        return {'integerValue': str(val)}
    if isinstance(val, float):
        return {'doubleValue': val}
    if isinstance(val, str):
        encoded = val.encode('utf-8')
        if len(encoded) > _MAX_STRING_LENGTH:
            return {
                'excludeFromIndexes': True,
                'blobValue': base64.b64encode(encoded)
                .replace(b'+', b'-')
                .replace(b'/', b'_')
                .rstrip(b'=')
                .decode('utf-8')}
        return {'stringValue': val}
    if isinstance(val, Key):
        return {'keyValue': val.get_dict()}
    if isinstance(val, list):
        return {'arrayValue': {'values': [value_to_dict(v) for v in val]}}
    if isinstance(val, TimestampValue):
        return {'timestampValue': str(val)}

    raise TypeError('Unsupported type: {}'.format(type(val)))


def value_from_dict(val):
    if 'nullValue' in val:
        return None  # val['nullValue'] is None
    if 'booleanValue' in val:
        return val['booleanValue']
    if 'integerValue' in val:
        return int(val['integerValue'])
    if 'doubleValue' in val:
        return val['doubleValue']
    if 'stringValue' in val:
        return val['stringValue']
    if 'keyValue' in val:
        return Key(val['keyValue'])
    if 'arrayValue' in val:
        return [
            value_from_dict(v)
            for v in val['arrayValue'].get('values', [])]
    if 'timestampValue' in val:
        return TimestampValue(val['timestampValue'])
    if 'blobValue' in val:
        val = val['blobValue'].encode('utf-8')
        val += b'=' * (4 - len(val) % 4)
        return base64.b64decode(
            val.replace(b'-', b'+').replace(b'_', b'/')).decode('utf-8')

    if 'entityValue' in val:
        #  TODO: Do something with entity values
        return None

    raise TypeError('Unexpected or unsupported value: {}'.format(val))



"""key.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import base64
from .buffer import Buffer
from .buffer import BufferDecodeError
from .path import Path
from .path import path_from_decoder
from .decoder import Decoder


class Key:
    KEY_INIT_MSG = """
        Key can be initialized by using a dictionary, for example:

            Key({
                "partitionId": {
                    "projectId": "my-project-id"
                },
                "path": [{
                  "kind": "Foo",
                  "id": "12345678"
                }]
            })

        Or by using a key string, for example:

            Key(ks="ag9zfm15LXByb2plY3QtaWRyDAsSA0ZvbxjOwvEFDA")

        Or, you can use arguments and set a project_id, for example:

            Key("Foo", 12345678, ..., project_id="my-project-id")

        Or you can use keyword arguments path and project_id, for example:

            Key(path=Path(...), project_id="my-project-id")
"""
    _ks = None

    def __init__(self, *args, ks=None, path=None, project_id=None):

        if len(args) == 1 and isinstance(args[0], dict):
            assert ks is None and path is None and project_id is None, \
                self.KEY_INIT_MSG

            res = args[0]
            self.project_id = res['partitionId']['projectId']
            self.path = Path(pairs=tuple(
                (pair['kind'], self._extract_id_or_name(pair))
                for pair in res['path']
            ))
            return

        if ks is not None:
            assert (not args) and path is None and project_id is None, \
                self.KEY_INIT_MSG
            self.project_id, self.path = self._deserialize_ks(ks)
            return

        assert project_id is not None and ((not args) ^ (path is None)), \
            self.KEY_INIT_MSG

        assert len(args) % 2 == 0, self.KEY_INIT_MSG

        self.project_id = project_id
        self.path = \
            Path(pairs=zip(*[iter(args)]*2)) if path is None else path

    def __repr__(self):
        return self.path.__repr__()

    def encode(self):
        """Return a Buffer() object which is a byte-like object.

        For serializing the key object you can use the .ks property which uses
        this method for generating an urlsafe key string.
        """
        buffer = Buffer()
        buffer.add_var_int32(106)

        # The project id in a key string is prefixed with s~
        buffer.add_prefixed_string('s~{}'.format(self.project_id))

        self.path.encode(buffer)
        return buffer

    @property
    def ks(self):
        if self._ks is None:
            self._ks = base64.b64encode(self.encode()) \
                    .rstrip(b'=') \
                    .replace(b'+', b'-') \
                    .replace(b'/', b'_') \
                    .decode('utf-8')
        return self._ks

    def get_dict(self):
        d = {'partitionId': {'projectId': self.project_id}}
        d.update(self.path.get_dict())
        return d

    @property
    def kind(self):
        """Shortcut for .path[-1].kind"""
        return self.path[-1].kind

    @property
    def id(self):
        """Shortcut for .path[-1].id"""
        return self.path[-1].id

    @staticmethod
    def _extract_id_or_name(pair):
        """Used on __init__."""
        if 'id' in pair:
            return int(pair['id'])

        if 'name' in pair:
            return pair['name']

        raise IndexError(
            'Both "id" and "name" are missing in pair: {}'.format(pair))

    @staticmethod
    def _deserialize_ks(ks):
        """Returns a Key() object from a key string."""

        decoder = Decoder(ks=ks)
        project_id = None
        path = None

        while decoder:
            tt = decoder.get_var_int32()

            if tt == 106:
                # The project id in a key string is prefixed with s~ which is
                # not part of the real project id.
                project_id = decoder.get_prefixed_string()[2:]
                continue

            if tt == 114:
                l = decoder.get_var_int32()
                decoder.set_end(l)
                path = path_from_decoder(decoder)
                decoder.set_end()

                continue

            if tt == 162:
                raise BufferDecodeError('namespaces are not supported')

            if tt == 0:
                raise BufferDecodeError('corrupt')

        return project_id, path



"""filter.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from ..connector.key import Key
from ..connector import GcdConnector


class Filter(dict):

    def __init__(self, model, *filters, has_ancestor=None, key=None):
        self._model = model
        filters = list(filters)

        if has_ancestor is not None:
            assert isinstance(has_ancestor, Key), \
                'Keyword argument \'has_ancestor\' should be of type ' \
                '\'Key\' but found type {!r}' \
                .format(has_ancestor.__class__.__name__)

            filters.append({
                'property': {'name': '__key__'},
                'value': {'keyValue': has_ancestor.get_dict()},
                'op': 'HAS_ANCESTOR'})

        if key is not None:
            assert isinstance(key, Key), \
                'Keyword argument \'key\' should be of type \'Key\' ' \
                'but found type {!r}' \
                .format(key.__class__.__name__)
            filters.append({
                'property': {'name': '__key__'},
                'value': {'keyValue': key.get_dict()},
                'op': 'EQUAL'})

        filter_dict = {'query': {'kind': [{'name': self._model.get_kind()}]}}

        if len(filters) == 1:
            filter_dict['query']['filter'] = {
                'propertyFilter': filters[0]
            }
        elif len(filters) > 1:
            filter_dict['query']['filter'] = {
                'compositeFilter': {
                    'op': 'AND',
                    'filters': [{'propertyFilter': f} for f in filters]
                }
            }

        super().__init__(**filter_dict)

    async def get_entity(self, gcd: GcdConnector):
        """Return a GcdModel instance from the supplied filter.

        :param gcd: GcdConnector instance.
        :return: GcdModel object or None in case no entity was found.
        """
        entity = await gcd.get_entity(self)
        return None if entity is None else self._model(entity)

    async def get_entities(self, gcd: GcdConnector) -> list:
        """Returns a list containing GcdModel instances from the supplied filter.

        :param gcd: GcdConnector instance.
        :return: list containing GcdModel objects.
        """
        return [self._model(ent) for ent in await gcd.get_entities(self)]

    async def get_key(self, gcd):
        return await gcd.get_key(self)

    async def get_keys(self, gcd):
        return await gcd.get_keys(self)
